#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import pytest

import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.signals as signals

from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import buy_limit_order

import octobot_trading.personal_data as personal_data


@pytest.fixture
def signal_builder():
    return signals.SignalBuilder(
        "strategy_name",
        "exchange_name",
        enums.ExchangeTypes.SPOT.value,
        "BTC/USDT",
        "signal description",
        enums.EvaluatorStates.SHORT.value,
        [],
    )


def test_build(signal_builder):
    trading_signal = signal_builder.build()
    assert trading_signal.strategy == "strategy_name"
    assert trading_signal.exchange == "exchange_name"
    assert trading_signal.exchange_type == enums.ExchangeTypes.SPOT.value
    assert trading_signal.symbol == "BTC/USDT"
    assert trading_signal.description == "signal description"
    assert trading_signal.state == enums.EvaluatorStates.SHORT.value
    assert trading_signal.orders == []

    signal_builder.orders = [None]
    assert signal_builder.build().orders == [None]


def test_reset(signal_builder):
    signal_builder.orders = ["hi"]
    signal_builder.reset()
    assert signal_builder.orders == []


def test_add_created_order(signal_builder, buy_limit_order):
    with pytest.raises(errors.InvalidArgumentError):
        signal_builder.add_created_order(buy_limit_order)
    signal_builder.add_created_order(buy_limit_order, target_amount="1%")
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.ACTION.value] \
        is enums.TradingSignalOrdersActions.CREATE.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.SHARED_SIGNAL_ORDER_ID.value] == \
           buy_limit_order.shared_signal_order_id
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] == "1%"
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None

    # add the same order: do not add it twice
    signal_builder.add_created_order(buy_limit_order, target_amount="1%")
    assert len(signal_builder.orders) == 1

    # update the same order
    buy_limit_order.order_type = enums.TraderOrderType.SELL_LIMIT
    signal_builder.add_created_order(buy_limit_order, target_position="2%")
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.SELL_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] == "2%"

    # add new order (orders are based on shared_signal_order_id)
    previous_shared_signal_order_id= buy_limit_order.shared_signal_order_id
    buy_limit_order.set_shared_signal_order_id("other_id")
    buy_limit_order.order_type = enums.TraderOrderType.STOP_LOSS_LIMIT
    signal_builder.add_created_order(buy_limit_order, target_position="50")
    assert len(signal_builder.orders) == 2
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.SHARED_SIGNAL_ORDER_ID.value] == \
           previous_shared_signal_order_id
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.SELL_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] == "2%"
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.SHARED_SIGNAL_ORDER_ID.value] == "other_id"
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] == "50"


def test_add_order_to_group(signal_builder, buy_limit_order):
    # no order_group
    signal_builder.add_order_to_group(buy_limit_order)
    # ensure properly added
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.ACTION.value] \
        is enums.TradingSignalOrdersActions.ADD_TO_GROUP.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.SHARED_SIGNAL_ORDER_ID.value] == \
           buy_limit_order.shared_signal_order_id
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None

    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_ID.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] is None

    order_group = personal_data.OneCancelsTheOtherOrderGroup(
        "group_name",
        buy_limit_order.exchange_manager.exchange_personal_data.orders_manager
    )
    buy_limit_order.add_to_order_group(order_group)
    signal_builder.add_order_to_group(buy_limit_order)
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_ID.value] == "group_name"
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == \
           personal_data.OneCancelsTheOtherOrderGroup.__name__

    # add the same order: do not add it twice
    signal_builder.add_order_to_group(buy_limit_order)
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_ID.value] == order_group.name
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == order_group.__class__.__name__

    # update the same order
    other_order_group = personal_data.BalancedTakeProfitAndStopOrderGroup(
        "group_name_2",
        buy_limit_order.exchange_manager.exchange_personal_data.orders_manager
    )
    buy_limit_order.add_to_order_group(other_order_group)
    signal_builder.add_order_to_group(buy_limit_order)
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_ID.value] == "group_name_2"
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == \
           personal_data.BalancedTakeProfitAndStopOrderGroup.__name__

    # add new order (orders are based on shared_signal_order_id)
    buy_limit_order.set_shared_signal_order_id("other_id")
    buy_limit_order.add_to_order_group(order_group)
    signal_builder.add_order_to_group(buy_limit_order)
    assert len(signal_builder.orders) == 2
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_ID.value] == "group_name_2"
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == \
           personal_data.BalancedTakeProfitAndStopOrderGroup.__name__
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.GROUP_ID.value] == "group_name"
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] == \
           personal_data.OneCancelsTheOtherOrderGroup.__name__


def test_add_edited_order(signal_builder, buy_limit_order):
    # no updated argument
    with pytest.raises(errors.InvalidArgumentError):
        signal_builder.add_edited_order(buy_limit_order)

    signal_builder.add_edited_order(buy_limit_order, updated_target_amount="1%")
    # ensure properly added
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.ACTION.value] \
        is enums.TradingSignalOrdersActions.EDIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.SHARED_SIGNAL_ORDER_ID.value] == \
           buy_limit_order.shared_signal_order_id
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] == "1%"
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] == "1%"
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 0.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 0.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0

    # add the same order: do not add it twice, update existing order
    signal_builder.add_edited_order(buy_limit_order, updated_target_position="1%")
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] == "1%"
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] == "1%"
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 0.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 0.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0
    signal_builder.add_edited_order(buy_limit_order, updated_limit_price=decimal.Decimal("1"))
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 1.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 0.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0
    signal_builder.add_edited_order(buy_limit_order, updated_stop_price=decimal.Decimal("1"))
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 0.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 1.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0
    signal_builder.add_edited_order(buy_limit_order, updated_current_price=decimal.Decimal("1"))
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] is None
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] == 0.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] == 0.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 1.0

    # add new order (orders are based on shared_signal_order_id)
    buy_limit_order.set_shared_signal_order_id("other_id")
    signal_builder.add_edited_order(buy_limit_order, updated_target_position="1%a")
    assert len(signal_builder.orders) == 2
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 1.0
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] == None
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] == 0.0
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] == "1%a"


def test_add_cancelled_order(signal_builder, buy_limit_order):
    signal_builder.add_cancelled_order(buy_limit_order)
    assert len(signal_builder.orders) == 1
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.ACTION.value] \
        is enums.TradingSignalOrdersActions.CANCEL.value
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.SHARED_SIGNAL_ORDER_ID.value] == \
           buy_limit_order.shared_signal_order_id

    # add the same order: do not add it twice
    signal_builder.add_cancelled_order(buy_limit_order)
    assert len(signal_builder.orders) == 1

    # add new order (orders are based on shared_signal_order_id)
    buy_limit_order.set_shared_signal_order_id("other_id")
    buy_limit_order.order_type = enums.TraderOrderType.STOP_LOSS_LIMIT
    signal_builder.add_created_order(buy_limit_order, target_position="50")
    assert len(signal_builder.orders) == 2
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value

    # add an orders via create action: it gets popped out of the orders list as there is no point creating
    # it and cancelling it right away
    buy_limit_order.set_shared_signal_order_id("buy_other_id")
    buy_limit_order.order_type = enums.TraderOrderType.BUY_MARKET
    signal_builder.add_created_order(buy_limit_order, target_amount="1")
    assert len(signal_builder.orders) == 3
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value

    # BUY_MARKET order got removed from orders
    signal_builder.add_cancelled_order(buy_limit_order)
    assert len(signal_builder.orders) == 2
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value

    # adding BUY_MARKET as add to group order
    other_order_group = personal_data.BalancedTakeProfitAndStopOrderGroup(
        "group_name_2",
        buy_limit_order.exchange_manager.exchange_personal_data.orders_manager
    )
    buy_limit_order.add_to_order_group(other_order_group)
    signal_builder.add_order_to_group(buy_limit_order)
    assert len(signal_builder.orders) == 3
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.ADD_TO_GROUP.value

    # BUY_MARKET order sill in order but action is not cancel from orders
    signal_builder.add_cancelled_order(buy_limit_order)
    assert len(signal_builder.orders) == 3
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.CANCEL.value

    # adding BUY_MARKET as edited order
    buy_limit_order.set_shared_signal_order_id("edit_buy_other_id")
    signal_builder.add_edited_order(buy_limit_order, updated_target_amount="1")
    assert len(signal_builder.orders) == 4
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.CANCEL.value
    assert signal_builder.orders[3][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert signal_builder.orders[3][enums.TradingSignalOrdersAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.EDIT.value

    # BUY_MARKET order sill in order but action is not cancel from orders
    signal_builder.add_cancelled_order(buy_limit_order)
    assert len(signal_builder.orders) == 4
    assert signal_builder.orders[0][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_LIMIT.value
    assert signal_builder.orders[1][enums.TradingSignalOrdersAttrs.TYPE.value] == \
           enums.TraderOrderType.STOP_LOSS_LIMIT.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert signal_builder.orders[2][enums.TradingSignalOrdersAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.CANCEL.value
    assert signal_builder.orders[3][enums.TradingSignalOrdersAttrs.TYPE.value] == enums.TraderOrderType.BUY_MARKET.value
    assert signal_builder.orders[3][enums.TradingSignalOrdersAttrs.ACTION.value] == \
           enums.TradingSignalOrdersActions.CANCEL.value



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import async_channel.channels as channels

import octobot_backtesting.api as api
import octobot_backtesting.enums as backtesting_enums

import octobot_commons.enums as common_enums
import octobot_commons.constants as common_constants
import octobot_commons.channels_name as channels_name
import octobot_commons.errors as errors

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.util as util
import octobot_trading.exchange_data.ticker.channel.ticker_updater as ticker_updater


class TickerUpdaterSimulator(ticker_updater.TickerUpdater):
    def __init__(self, channel, importer):
        super().__init__(channel)
        self.exchange_data_importer = importer
        self.exchange_name = self.channel.exchange_manager.exchange_name

        self.last_timestamp_pushed = 0
        self.last_timestamp_pushed_by_symbol = {}
        self.time_consumer = None
        # Only generate tickers from the shortest handled time frame
        shortest_time_frame = self.channel.exchange_manager.exchange_config.get_shortest_time_frame()
        self.ticker_time_frame = shortest_time_frame.value
        self.ticker_time_frame_seconds = common_enums.TimeFramesMinutes[shortest_time_frame] * \
            common_constants.MINUTE_TO_SECONDS

    async def start(self):
        await self.resume()

    async def handle_timestamp(self, timestamp, **kwargs):
        try:
            for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                ticker_data = (await self.exchange_data_importer.get_ticker_from_timestamps(
                    exchange_name=self.exchange_name,
                    symbol=pair,
                    inferior_timestamp=timestamp,
                    limit=1))[0]
                if ticker_data[0] > self.last_timestamp_pushed:
                    self.last_timestamp_pushed = ticker_data[0]
                    await self.push(pair, ticker_data[-1])
        except errors.DatabaseNotFoundError as e:
            self.logger.warning(f"Not enough data : {e}")
            await self.pause()
            await self.stop()
        except IndexError as e:
            self.logger.warning(f"Failed to access ticker_data : {e}")

    async def _ticker_from_ohlcv_callback(self, exchange: str, exchange_id: str,
                                          cryptocurrency: str, symbol: str, time_frame, candle):
        if self.ticker_time_frame == time_frame and candle:
            last_candle_timestamp = candle[common_enums.PriceIndexes.IND_PRICE_TIME.value]
            ticker_timestamp = last_candle_timestamp + self.ticker_time_frame_seconds
            if ticker_timestamp > self.last_timestamp_pushed_by_symbol[symbol]:
                self.last_timestamp_pushed_by_symbol[symbol] = ticker_timestamp
                ticker = self._generate_ticker_from_candle(candle, symbol, ticker_timestamp)
                await self.push(symbol, ticker)

    @staticmethod
    def _generate_ticker_from_candle(candle, symbol, last_candle_timestamp):
        return {
            enums.ExchangeConstantsTickersColumns.TIMESTAMP.value: last_candle_timestamp,
            enums.ExchangeConstantsTickersColumns.LAST.value: candle[common_enums.PriceIndexes.IND_PRICE_CLOSE.value],
            enums.ExchangeConstantsTickersColumns.SYMBOL.value: symbol,
            enums.ExchangeConstantsTickersColumns.HIGH.value: candle[common_enums.PriceIndexes.IND_PRICE_HIGH.value],
            enums.ExchangeConstantsTickersColumns.LOW.value: candle[common_enums.PriceIndexes.IND_PRICE_LOW.value],
            enums.ExchangeConstantsTickersColumns.OPEN.value: candle[common_enums.PriceIndexes.IND_PRICE_OPEN.value],
            enums.ExchangeConstantsTickersColumns.CLOSE.value: candle[common_enums.PriceIndexes.IND_PRICE_CLOSE.value]
        }

    async def pause(self):
        if self.time_consumer is not None:
            await channels.get_chan(channels_name.OctoBotBacktestingChannelsName.TIME_CHANNEL.value). \
                remove_consumer(self.time_consumer)
        self.is_running = False

    async def stop(self):
        await util.stop_and_pause(self)

    async def resume(self):
        if not self.is_running:
            if self.time_consumer is None and not self.channel.is_paused:
                if backtesting_enums.ExchangeDataTables.TICKER in \
                        api.get_available_data_types(self.exchange_data_importer):
                    self.time_consumer = await channels.get_chan(
                        channels_name.OctoBotBacktestingChannelsName.TIME_CHANNEL.value).new_consumer(
                        self.handle_timestamp)
                else:
                    await exchanges_channel.get_chan(constants.OHLCV_CHANNEL, self.channel.exchange_manager.id) \
                        .new_consumer(self._ticker_from_ohlcv_callback)
                    self.last_timestamp_pushed_by_symbol = {
                        symbol: 0
                        for symbol in self.channel.exchange_manager.exchange_config.traded_symbol_pairs
                    }
                self.is_running = True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import octobot_trading.enums as enums
import octobot_trading.personal_data as personal_data

from tests import event_loop
from tests.exchanges import future_simulated_exchange_manager
from tests.exchanges.traders import future_trader, future_trader_simulator_with_default_linear, DEFAULT_FUTURE_SYMBOL, \
    DEFAULT_FUTURE_SYMBOL_MARGIN_TYPE, DEFAULT_FUTURE_SYMBOL_LEVERAGE, get_default_future_inverse_contract
from tests.test_utils.random_numbers import decimal_random_int, decimal_random_quantity

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_create_position_instance_from_raw(future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear

    raw_position = {
        enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL,
        enums.ExchangeConstantsPositionColumns.STATUS.value: enums.PositionStatus.ADL.value
    }
    position = personal_data.create_position_instance_from_raw(trader_inst, raw_position)
    position_leverage = decimal_random_int(min_value=2, max_value=200)
    position_quantity = decimal_random_quantity(max_value=1000)
    exchange_manager_inst.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL,
                                                            get_default_future_inverse_contract())
    inverse_position_open = personal_data.create_position_instance_from_raw(trader_inst, {
        enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL,
        enums.ExchangeConstantsPositionColumns.QUANTITY.value: position_quantity
    })
    assert position.symbol == DEFAULT_FUTURE_SYMBOL
    assert position.market == "USDT"
    assert position.currency == "BTC"
    assert position.status == enums.PositionStatus.ADL
    assert isinstance(position, personal_data.LinearPosition)

    assert inverse_position_open.status == enums.PositionStatus.OPEN
    assert inverse_position_open.quantity == position_quantity
    assert isinstance(inverse_position_open, personal_data.InversePosition)


async def test_create_symbol_position(future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    position = personal_data.create_symbol_position(trader_inst, DEFAULT_FUTURE_SYMBOL)
    assert position.symbol == DEFAULT_FUTURE_SYMBOL



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import os
import decimal

import pytest
import octobot_commons.constants as commons_constants

import octobot_trading.constants as constants
import octobot_trading.errors as errors
from octobot_trading.enums import TraderOrderType, TradeOrderSide
from octobot_trading.personal_data.orders import BuyLimitOrder
from octobot_trading.personal_data.orders import SellLimitOrder
from octobot_trading.personal_data.orders import StopLossOrder
from octobot_trading.personal_data.orders import BuyMarketOrder
from octobot_trading.personal_data.orders.types.market.sell_market_order import SellMarketOrder
import octobot_trading.personal_data.orders.groups as order_groups
from tests.test_utils.order_util import fill_market_order, fill_limit_or_stop_order

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_load_portfolio(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    assert portfolio_manager.portfolio.portfolio['BTC'].available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.portfolio['BTC'].total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.portfolio['USDT'].available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.portfolio['USDT'].total == decimal.Decimal('1000')


async def test_copy_portfolio(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    pf1 = portfolio_manager.portfolio
    pf2 = copy.copy(pf1)
    pf1.portfolio['BTC'].available = decimal.Decimal('5')
    pf1.portfolio['BTC'].total = decimal.Decimal('9')
    pf1.portfolio['USDT'].available = decimal.Decimal('10.5')
    pf1.portfolio['USDT'].total = decimal.Decimal('1.3')
    assert pf1.portfolio['BTC'].available != pf2.portfolio['BTC'].available
    assert pf1.portfolio['BTC'].total != pf2.portfolio['BTC'].total
    assert pf1.portfolio['USDT'].available != pf2.portfolio['USDT'].available
    assert pf1.portfolio['USDT'].total != pf2.portfolio['USDT'].total
    assert pf1 != pf2


async def test_get_portfolio_from_amount_dict(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    assert portfolio_manager.portfolio.get_portfolio_from_amount_dict(
        {"zyx": decimal.Decimal(str(10)), "BTC": decimal.Decimal('1')}) == {
               'zyx': {'available': decimal.Decimal('10'), 'total': decimal.Decimal('10')},
               'BTC': {'available': decimal.Decimal('1'), 'total': decimal.Decimal('1')}
           }
    assert portfolio_manager.portfolio.get_portfolio_from_amount_dict({}) == {}
    with pytest.raises(RuntimeError):
        portfolio_manager.portfolio.get_portfolio_from_amount_dict({"zyx": "10", "BTC": decimal.Decimal('1')})
    with pytest.raises(RuntimeError):
        portfolio_manager.portfolio.get_portfolio_from_amount_dict({"BTC": 1})


async def test_get_currency_portfolio(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("NANO").total == decimal.Decimal('0')


async def test_update_portfolio_from_balance(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    test_portfolio = {"zyx": {commons_constants.PORTFOLIO_AVAILABLE: decimal.Decimal('0.1'),
                              commons_constants.PORTFOLIO_TOTAL: decimal.Decimal('0.1')},
                      "BTC": {commons_constants.PORTFOLIO_AVAILABLE: decimal.Decimal('1'),
                              commons_constants.PORTFOLIO_TOTAL: decimal.Decimal('1')},
                      "ETH": {commons_constants.PORTFOLIO_AVAILABLE: decimal.Decimal('50'),
                              commons_constants.PORTFOLIO_TOTAL: decimal.Decimal('150')}}

    assert portfolio_manager.portfolio.update_portfolio_from_balance(test_portfolio)
    assert portfolio_manager.portfolio.portfolio['BTC'].available == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['BTC'].total == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['zyx'].available == decimal.Decimal('0.1')
    assert portfolio_manager.portfolio.portfolio['zyx'].total == decimal.Decimal('0.1')
    assert portfolio_manager.portfolio.portfolio['ETH'].available == decimal.Decimal('50')
    assert portfolio_manager.portfolio.portfolio['ETH'].total == decimal.Decimal('150')

    # should return False when no update made
    assert not portfolio_manager.portfolio.update_portfolio_from_balance(test_portfolio, force_replace=False)


async def test_update_portfolio_from_balance_with_deltas(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    test_portfolio = {"BTC": {commons_constants.PORTFOLIO_AVAILABLE: decimal.Decimal('1'),
                              commons_constants.PORTFOLIO_TOTAL: decimal.Decimal('1')}}

    assert portfolio_manager.portfolio.update_portfolio_from_balance(test_portfolio, force_replace=False)
    assert portfolio_manager.portfolio.portfolio['BTC'].available == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['BTC'].total == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['USDT'].available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.portfolio['USDT'].total == decimal.Decimal('1000')

    test_portfolio_2 = {"NANO": {commons_constants.PORTFOLIO_AVAILABLE: decimal.Decimal('1'),
                                 commons_constants.PORTFOLIO_TOTAL: decimal.Decimal('2')}}

    assert portfolio_manager.portfolio.update_portfolio_from_balance(test_portfolio_2, force_replace=False)
    assert portfolio_manager.portfolio.portfolio['BTC'].available == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['BTC'].total == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['USDT'].available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.portfolio['USDT'].total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.portfolio['NANO'].available == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['NANO'].total == decimal.Decimal('2')

    test_portfolio_3 = {"USDT": {commons_constants.PORTFOLIO_AVAILABLE: decimal.Decimal('250'),
                                 commons_constants.PORTFOLIO_TOTAL: decimal.Decimal('500')}}

    assert portfolio_manager.portfolio.update_portfolio_from_balance(test_portfolio_3, force_replace=False)
    assert portfolio_manager.portfolio.portfolio['BTC'].available == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['BTC'].total == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['USDT'].available == decimal.Decimal('250')
    assert portfolio_manager.portfolio.portfolio['USDT'].total == decimal.Decimal('500')
    assert portfolio_manager.portfolio.portfolio['NANO'].available == decimal.Decimal('1')
    assert portfolio_manager.portfolio.portfolio['NANO'].total == decimal.Decimal('2')

    test_portfolio_4 = {"USDT": {commons_constants.PORTFOLIO_AVAILABLE: decimal.Decimal('100'),
                                 commons_constants.PORTFOLIO_TOTAL: decimal.Decimal('100')}}

    assert portfolio_manager.portfolio.update_portfolio_from_balance(test_portfolio_4, force_replace=True)
    assert portfolio_manager.portfolio.portfolio['USDT'].available == decimal.Decimal('100')
    assert portfolio_manager.portfolio.portfolio['USDT'].total == decimal.Decimal('100')


async def test_update_portfolio(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="BTC/USDT",
                     current_price=70,
                     quantity=10,
                     price=70)

    # update portfolio with creations
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('300')

    await fill_limit_or_stop_order(limit_buy)

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('20')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('300')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('20')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('300')

    # Test buy order
    market_sell = SellMarketOrder(trader)
    market_sell.update(order_type=TraderOrderType.SELL_MARKET,
                       symbol="BTC/USDT",
                       current_price=80,
                       quantity=8,
                       price=80)

    # update portfolio with creations
    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('12')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('300')

    await fill_market_order(market_sell)

    # when filling market sell
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('12')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('940')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('12')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('940')


async def test_update_portfolio_with_filled_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # force fees => should have consequences
    exchange_manager.exchange.config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_SIMULATOR_FEES] = {
        commons_constants.CONFIG_SIMULATOR_FEES_MAKER: 0.05,
        commons_constants.CONFIG_SIMULATOR_FEES_TAKER: 0.1
    }

    # Test buy order
    market_sell = SellMarketOrder(trader)
    market_sell.update(order_type=TraderOrderType.SELL_MARKET,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(70)),
                       quantity=decimal.Decimal(str(3)),
                       price=decimal.Decimal(str(70)))

    # Test sell order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal(str(100)),
                      quantity=decimal.Decimal(str(4.2)),
                      price=decimal.Decimal(str(100)))

    # Test stop loss order
    stop_loss = StopLossOrder(trader)
    stop_loss.update(order_type=TraderOrderType.STOP_LOSS,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(80)),
                     quantity=decimal.Decimal(str(4.2)),
                     price=decimal.Decimal(str(80)))

    oco_group = exchange_manager.exchange_personal_data.orders_manager\
        .create_group(order_groups.OneCancelsTheOtherOrderGroup)
    limit_sell.add_to_order_group(oco_group)
    stop_loss.add_to_order_group(oco_group)

    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(50)),
                     quantity=decimal.Decimal(str(2)),
                     price=decimal.Decimal(str(50)))

    # update portfolio with creations
    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    portfolio_manager.portfolio.update_portfolio_available(stop_loss, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('2.8')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('900')

    # when cancelling limit sell, market sell and stop orders
    portfolio_manager.portfolio.update_portfolio_available(stop_loss, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, False)

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('7')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('900')

    # when filling limit buy
    await fill_limit_or_stop_order(limit_buy)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('8.999')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('900')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('11.999')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('900')

    # when filling market sell
    await fill_market_order(market_sell)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '8.999')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '1109.79')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '8.999')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1109.79')


async def test_update_portfolio_with_cancelled_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # force fees => shouldn't do anything
    exchange_manager.exchange.config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_SIMULATOR_FEES] = {
        commons_constants.CONFIG_SIMULATOR_FEES_MAKER: decimal.Decimal('0.05'),
        commons_constants.CONFIG_SIMULATOR_FEES_TAKER: decimal.Decimal('0.1')
    }

    # Test buy order
    market_sell = SellMarketOrder(trader)
    market_sell.update(order_type=TraderOrderType.SELL_MARKET,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(80)),
                       quantity=decimal.Decimal(str(4.1)),
                       price=decimal.Decimal(str(80)))

    # Test sell order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal(str(10)),
                      quantity=decimal.Decimal(str(4.2)),
                      price=decimal.Decimal(str(10)))

    # Test stop loss order
    stop_loss = StopLossOrder(trader)
    stop_loss.update(order_type=TraderOrderType.STOP_LOSS,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(80)),
                     quantity=decimal.Decimal(str(3.6)),
                     price=decimal.Decimal(str(80)))

    portfolio_manager.portfolio.update_portfolio_available(stop_loss, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)

    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(50)),
                     quantity=decimal.Decimal(str(4)),
                     price=decimal.Decimal(str(50)))

    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)
    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)

    assert round(portfolio_manager.portfolio.get_currency_portfolio("BTC").available,
                 1) == decimal.Decimal('1.7')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '800')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1000')

    # with no filled orders
    portfolio_manager.portfolio.update_portfolio_available(stop_loss, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, False)
    portfolio_manager.portfolio.update_portfolio_available(market_sell, False)

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1000')


async def test_update_portfolio_with_stop_loss_sell_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal(str(90)),
                      quantity=decimal.Decimal(str(4)),
                      price=decimal.Decimal(str(90)))

    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(50)),
                     quantity=decimal.Decimal(str(4)),
                     price=decimal.Decimal(str(50)))

    # Test stop loss order
    stop_loss = StopLossOrder(trader, side=TradeOrderSide.SELL)
    stop_loss.update(order_type=TraderOrderType.STOP_LOSS,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(60)),
                     quantity=decimal.Decimal(str(4)),
                     price=decimal.Decimal(str(60)))

    portfolio_manager.portfolio.update_portfolio_available(stop_loss, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)

    assert round(portfolio_manager.portfolio.get_currency_portfolio("BTC").available, 1) == decimal.Decimal('6')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('800')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # cancel limits
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, False)

    await fill_limit_or_stop_order(stop_loss)

    # filling stop loss
    # typical stop loss behavior --> update available before update portfolio
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('6')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1240')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('6')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1240')


async def test_update_portfolio_with_stop_loss_buy_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal(str(90)),
                      quantity=decimal.Decimal(str(4)),
                      price=decimal.Decimal(str(90)))

    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(50)),
                     quantity=decimal.Decimal(str(4)),
                     price=decimal.Decimal(str(50)))

    # Test stop loss order
    stop_loss = StopLossOrder(trader, side=TradeOrderSide.BUY)
    stop_loss.update(order_type=TraderOrderType.STOP_LOSS,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(60)),
                     quantity=decimal.Decimal(str(4)),
                     price=decimal.Decimal(str(60)))

    portfolio_manager.portfolio.update_portfolio_available(stop_loss, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)

    assert round(portfolio_manager.portfolio.get_currency_portfolio("BTC").available, 1) == decimal.Decimal('6')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('800')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # cancel limits
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, False)

    await fill_limit_or_stop_order(stop_loss)

    # filling stop loss
    # typical stop loss behavior --> update available before update portfolio
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('14')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('760')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('14')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('760')


async def test_update_portfolio_with_some_filled_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal(str(90)),
                      quantity=decimal.Decimal(str(4)),
                      price=decimal.Decimal(str(90)))

    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(60)),
                     quantity=decimal.Decimal(str(2)),
                     price=decimal.Decimal(str(60)))

    # Test buy order
    limit_buy_2 = BuyLimitOrder(trader)
    limit_buy_2.update(order_type=TraderOrderType.BUY_LIMIT,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(50)),
                       quantity=decimal.Decimal(str(4)),
                       price=decimal.Decimal(str(50)))

    # Test sell order
    limit_sell_2 = SellLimitOrder(trader)
    limit_sell_2.update(order_type=TraderOrderType.SELL_LIMIT,
                        symbol="BTC/USDT",
                        current_price=decimal.Decimal(str(10)),
                        quantity=decimal.Decimal(str(2)),
                        price=decimal.Decimal(str(10)))

    # Test stop loss order
    stop_loss_2 = StopLossOrder(trader)
    stop_loss_2.update(order_type=TraderOrderType.STOP_LOSS,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(10)),
                       quantity=decimal.Decimal(str(2)),
                       price=decimal.Decimal(str(10)))

    # Test sell order
    limit_sell_3 = SellLimitOrder(trader)
    limit_sell_3.update(order_type=TraderOrderType.SELL_LIMIT,
                        symbol="BTC/USDT",
                        current_price=decimal.Decimal(str(20)),
                        quantity=decimal.Decimal(str(1)),
                        price=decimal.Decimal(str(20)))

    # Test stop loss order
    stop_loss_3 = StopLossOrder(trader)
    stop_loss_3.update(order_type=TraderOrderType.STOP_LOSS,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(20)),
                       quantity=decimal.Decimal(str(1)),
                       price=decimal.Decimal(str(20)))

    portfolio_manager.portfolio.update_portfolio_available(stop_loss_2, True)
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_3, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_2, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_3, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy_2, True)

    assert round(portfolio_manager.portfolio.get_currency_portfolio("BTC").available,
                 1) == 3
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '680')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1000')

    # Test stop loss order
    stop_loss_4 = StopLossOrder(trader, side=TradeOrderSide.BUY)
    stop_loss_4.update(order_type=TraderOrderType.STOP_LOSS,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(20)),
                       quantity=decimal.Decimal(str(4)),
                       price=decimal.Decimal(str(20)))

    # Test stop loss order
    stop_loss_5 = StopLossOrder(trader, side=TradeOrderSide.BUY)
    stop_loss_5.update(order_type=TraderOrderType.STOP_LOSS,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(200)),
                       quantity=decimal.Decimal(str(4)),
                       price=decimal.Decimal(str(20)))
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_5, True)

    # portfolio did not change as stop losses are not affecting available funds
    assert round(portfolio_manager.portfolio.get_currency_portfolio("BTC").available,
                 1) == 3
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '680')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1000')

    # cancels
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_3, False)
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_5, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_2, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, False)

    # filling
    await fill_limit_or_stop_order(stop_loss_2)
    await fill_limit_or_stop_order(limit_sell)
    await fill_limit_or_stop_order(limit_sell_3)
    await fill_limit_or_stop_order(limit_buy_2)

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '7')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '1200')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('7')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1200')

    await fill_limit_or_stop_order(stop_loss_4)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '11')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '1120')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '11')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1120')


async def test_update_portfolio_with_multiple_filled_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal(str(90)),
                      quantity=decimal.Decimal(str(4)),
                      price=decimal.Decimal(str(90)))

    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="BTC/USDT",
                     current_price=decimal.Decimal(str(60)),
                     quantity=decimal.Decimal(str(2)),
                     price=decimal.Decimal(str(60)))

    # Test buy order
    limit_buy_2 = BuyLimitOrder(trader)
    limit_buy_2.update(order_type=TraderOrderType.BUY_LIMIT,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(50)),
                       quantity=decimal.Decimal(str(4)),
                       price=decimal.Decimal(str(50)))

    # Test buy order
    limit_buy_3 = BuyLimitOrder(trader)
    limit_buy_3.update(order_type=TraderOrderType.BUY_LIMIT,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(46)),
                       quantity=decimal.Decimal(str(2)),
                       price=decimal.Decimal(str(46)))

    # Test buy order
    limit_buy_4 = BuyLimitOrder(trader)
    limit_buy_4.update(order_type=TraderOrderType.BUY_LIMIT,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(41)),
                       quantity=decimal.Decimal(str(1.78)),
                       price=decimal.Decimal(str(41)))

    # Test buy order
    limit_buy_5 = BuyLimitOrder(trader)
    limit_buy_5.update(order_type=TraderOrderType.BUY_LIMIT,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(0.2122427)),
                       quantity=decimal.Decimal(str(3.72448)),
                       price=decimal.Decimal(str(0.2122427)))

    # Test buy order
    limit_buy_6 = BuyLimitOrder(trader)
    limit_buy_6.update(order_type=TraderOrderType.BUY_LIMIT,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(430)),
                       quantity=decimal.Decimal(str(1.05)),
                       price=decimal.Decimal(str(430)))

    # Test sell order
    limit_sell_2 = SellLimitOrder(trader)
    limit_sell_2.update(order_type=TraderOrderType.SELL_LIMIT,
                        symbol="BTC/USDT",
                        current_price=decimal.Decimal(str(10)),
                        quantity=decimal.Decimal(str(2)),
                        price=decimal.Decimal(str(10)))

    # Test stop loss order
    stop_loss_2 = StopLossOrder(trader)
    stop_loss_2.update(order_type=TraderOrderType.STOP_LOSS,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(10)),
                       quantity=decimal.Decimal(str(2)),
                       price=decimal.Decimal(str(10)))

    # Test sell order
    limit_sell_3 = SellLimitOrder(trader)
    limit_sell_3.update(order_type=TraderOrderType.SELL_LIMIT,
                        symbol="BTC/USDT",
                        current_price=decimal.Decimal(str(20)),
                        quantity=decimal.Decimal(str(1)),
                        price=decimal.Decimal(str(20)))

    # Test stop loss order
    stop_loss_3 = StopLossOrder(trader)
    stop_loss_3.update(order_type=TraderOrderType.STOP_LOSS,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(20)),
                       quantity=decimal.Decimal(str(1)),
                       price=decimal.Decimal(str(20)))

    # Test sell order
    limit_sell_4 = SellLimitOrder(trader)
    limit_sell_4.update(order_type=TraderOrderType.SELL_LIMIT,
                        symbol="BTC/USDT",
                        current_price=decimal.Decimal(str(50)),
                        quantity=decimal.Decimal(str(0.2)),
                        price=decimal.Decimal(str(50)))

    # Test stop loss order
    stop_loss_4 = StopLossOrder(trader, side=TradeOrderSide.BUY)
    stop_loss_4.update(order_type=TraderOrderType.STOP_LOSS,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(45)),
                       quantity=decimal.Decimal(str(0.2)),
                       price=decimal.Decimal(str(45)))

    # Test sell order
    limit_sell_5 = SellLimitOrder(trader)
    limit_sell_5.update(order_type=TraderOrderType.SELL_LIMIT,
                        symbol="BTC/USDT",
                        current_price=decimal.Decimal(str(11)),
                        quantity=decimal.Decimal(str(0.7)),
                        price=decimal.Decimal(str(11)))

    # Test stop loss order
    stop_loss_5 = StopLossOrder(trader)
    stop_loss_5.update(order_type=TraderOrderType.STOP_LOSS,
                       symbol="BTC/USDT",
                       current_price=decimal.Decimal(str(9)),
                       quantity=decimal.Decimal(str(0.7)),
                       price=decimal.Decimal(str(9)))

    portfolio_manager.portfolio.update_portfolio_available(stop_loss_2, True)
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_3, True)
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_4, True)
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_5, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_2, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_3, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_4, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_5, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy_2, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy_3, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy_4, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy_5, True)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy_6, True)

    assert round(portfolio_manager.portfolio.get_currency_portfolio("BTC").available,
                 1) == decimal.Decimal('2.1')
    assert round(portfolio_manager.portfolio.get_currency_portfolio("USDT").available,
                 7) == decimal.Decimal('62.7295063')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1000')

    # cancels
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_3, False)
    portfolio_manager.portfolio.update_portfolio_available(stop_loss_5, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_2, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy_3, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_buy_5, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell_4, False)

    # filling
    await fill_limit_or_stop_order(stop_loss_2)
    await fill_limit_or_stop_order(limit_sell)
    await fill_limit_or_stop_order(limit_sell_3)
    await fill_limit_or_stop_order(limit_buy_2)
    await fill_limit_or_stop_order(limit_sell_5)
    await fill_limit_or_stop_order(stop_loss_4)
    await fill_limit_or_stop_order(limit_buy_4)
    await fill_limit_or_stop_order(limit_buy_5)
    await fill_limit_or_stop_order(limit_buy_6)

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '13.05448')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '674.22')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '13.05448')
    assert round(portfolio_manager.portfolio.get_currency_portfolio("USDT").total,
                 7) == decimal.Decimal('673.4295063')


async def test_update_portfolio_with_multiple_symbols_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    market_buy = BuyMarketOrder(trader)
    market_buy.update(order_type=TraderOrderType.BUY_MARKET,
                      symbol="ETH/USDT",
                      current_price=decimal.Decimal(str(7)),
                      quantity=decimal.Decimal(str(100)),
                      price=decimal.Decimal(str(7)))

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("ETH").available == decimal.Decimal('0')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('300')
    assert portfolio_manager.portfolio.get_currency_portfolio("ETH").total == decimal.Decimal('0')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    await fill_market_order(market_buy)

    assert portfolio_manager.portfolio.get_currency_portfolio("ETH").available == decimal.Decimal('100')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('300')
    assert portfolio_manager.portfolio.get_currency_portfolio("ETH").total == decimal.Decimal('100')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('300')

    # Test buy order
    market_buy = BuyMarketOrder(trader)
    market_buy.update(order_type=TraderOrderType.BUY_MARKET,
                      symbol="LTC/BTC",
                      current_price=decimal.Decimal(str(0.0135222)),
                      quantity=decimal.Decimal(str(100)),
                      price=decimal.Decimal(str(0.0135222)))

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("LTC").available == decimal.Decimal('0')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('8.64778')
    assert portfolio_manager.portfolio.get_currency_portfolio("LTC").total == decimal.Decimal('0')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')

    await fill_market_order(market_buy)

    assert portfolio_manager.portfolio.get_currency_portfolio("LTC").available == decimal.Decimal('100')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('8.64778')
    assert portfolio_manager.portfolio.get_currency_portfolio("LTC").total == decimal.Decimal('100')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('8.64778')

    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="XRP/BTC",
                     current_price=decimal.Decimal(str(0.00012232132312312)),
                     quantity=decimal.Decimal(str(3000.1214545)),
                     price=decimal.Decimal(str(0.00012232132312312)))

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("XRP").available == decimal.Decimal('0')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('8.280801174155500743021960')
    assert portfolio_manager.portfolio.get_currency_portfolio("XRP").total == decimal.Decimal('0')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('8.64778')

    # cancel
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("XRP").available == decimal.Decimal('0')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('8.64778')
    assert portfolio_manager.portfolio.get_currency_portfolio("XRP").total == decimal.Decimal('0')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('8.64778')


async def test_reset_portfolio_available(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal(str(90)),
                      quantity=decimal.Decimal(str(4)),
                      price=decimal.Decimal(str(90)))

    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    portfolio_manager.portfolio.reset_portfolio_available()

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1000')

    # Test sell order
    limit_sell = SellLimitOrder(trader)
    limit_sell.update(order_type=TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal(str(90)),
                      quantity=decimal.Decimal(str(4)),
                      price=decimal.Decimal(str(90)))

    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    # Test buy order
    limit_buy = BuyLimitOrder(trader)
    limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                     symbol="VEN/BTC",
                     current_price=decimal.Decimal(str(0.5)),
                     quantity=decimal.Decimal(str(4)),
                     price=decimal.Decimal(str(0.5)))

    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)

    # Test buy order
    btc_limit_buy = BuyLimitOrder(trader)
    btc_limit_buy.update(order_type=TraderOrderType.BUY_LIMIT,
                         symbol="BTC/USDT",
                         current_price=decimal.Decimal(str(10)),
                         quantity=decimal.Decimal(str(50)),
                         price=decimal.Decimal(str(10)))

    portfolio_manager.portfolio.update_portfolio_available(btc_limit_buy, True)

    # Test buy order
    btc_limit_buy2 = BuyLimitOrder(trader)
    btc_limit_buy2.update(order_type=TraderOrderType.BUY_LIMIT,
                          symbol="BTC/USDT",
                          current_price=decimal.Decimal(str(10)),
                          quantity=decimal.Decimal(str(50)),
                          price=decimal.Decimal(str(10)))

    portfolio_manager.portfolio.update_portfolio_available(btc_limit_buy2, True)

    # reset equivalent of the ven buy order
    portfolio_manager.portfolio.reset_portfolio_available("BTC", decimal.Decimal(str(4 * 0.5)))

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '6')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '0')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1000')

    # reset equivalent of the btc buy orders 1 and 2
    portfolio_manager.portfolio.reset_portfolio_available("USDT")

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '6')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(
        '1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(
        '1000')


async def test_default_impl(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    order = BuyMarketOrder(trader)

    # should not raise NotImplemented
    portfolio_manager.portfolio.update_portfolio_data_from_order(order)
    portfolio_manager.portfolio.update_portfolio_available_from_order(order)
    portfolio_manager.portfolio.log_portfolio_update_from_order(order)


async def test_parse_currency_balance(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    if not os.getenv('CYTHON_IGNORE'):
        # 0 values
        parsed_portfolio = portfolio_manager.portfolio._parse_raw_currency_asset("BTC",
            {commons_constants.PORTFOLIO_AVAILABLE: constants.ZERO,
             commons_constants.PORTFOLIO_TOTAL: constants.ZERO})
        assert parsed_portfolio.available == constants.ZERO
        assert parsed_portfolio.total == constants.ZERO

        parsed_portfolio_2 = portfolio_manager.portfolio._parse_raw_currency_asset("BTC",
            {constants.CONFIG_PORTFOLIO_FREE: constants.ZERO,
             commons_constants.PORTFOLIO_TOTAL: constants.ZERO})
        assert parsed_portfolio_2.available == constants.ZERO
        assert parsed_portfolio_2.total == constants.ZERO

        # None values
        parsed_portfolio_3 = portfolio_manager.portfolio._parse_raw_currency_asset("BTC",
            {commons_constants.PORTFOLIO_AVAILABLE: None,
             commons_constants.PORTFOLIO_TOTAL: constants.ZERO})
        assert parsed_portfolio_3.available == constants.ZERO
        assert parsed_portfolio_3.total == constants.ZERO

        parsed_portfolio_4 = portfolio_manager.portfolio._parse_raw_currency_asset("BTC",
            {commons_constants.PORTFOLIO_AVAILABLE: None,
             commons_constants.PORTFOLIO_TOTAL: None})
        assert parsed_portfolio_4.available == constants.ZERO
        assert parsed_portfolio_4.total == constants.ZERO

        parsed_portfolio_5 = portfolio_manager.portfolio._parse_raw_currency_asset("BTC",
            {commons_constants.PORTFOLIO_AVAILABLE: constants.ZERO,
             commons_constants.PORTFOLIO_TOTAL: None})
        assert parsed_portfolio_5.available == constants.ZERO
        assert parsed_portfolio_5.total == constants.ZERO

        parsed_portfolio_6 = portfolio_manager.portfolio._parse_raw_currency_asset("BTC",
             {constants.CONFIG_PORTFOLIO_FREE: None,
              commons_constants.PORTFOLIO_TOTAL: 0.0})
        assert parsed_portfolio_6.available == constants.ZERO
        assert parsed_portfolio_6.total == constants.ZERO

        parsed_portfolio_7 = portfolio_manager.portfolio._parse_raw_currency_asset("BTC",
             {constants.CONFIG_PORTFOLIO_FREE: None,
             commons_constants.PORTFOLIO_TOTAL: None})
        assert parsed_portfolio_7.available == constants.ZERO
        assert parsed_portfolio_7.total == constants.ZERO

        parsed_portfolio_8 = portfolio_manager.portfolio._parse_raw_currency_asset("BTC",
            {constants.CONFIG_PORTFOLIO_FREE: constants.ZERO,
             commons_constants.PORTFOLIO_TOTAL: None})
        assert parsed_portfolio_8.available == constants.ZERO
        assert parsed_portfolio_8.total == constants.ZERO


async def test_update_portfolio_data(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    if not os.getenv('CYTHON_IGNORE'):
        # doesn't raise
        portfolio_manager.portfolio._update_portfolio_data("BTC", -1)

        with pytest.raises(errors.PortfolioNegativeValueError):
            portfolio_manager.portfolio._update_portfolio_data("USDT", decimal.Decimal(-2000))
        with pytest.raises(errors.PortfolioNegativeValueError):
            portfolio_manager.portfolio._update_portfolio_data("BTC", decimal.Decimal(-20))

    if not os.getenv('CYTHON_IGNORE'):
        with pytest.raises(errors.PortfolioNegativeValueError):
            # Test buy order
            btc_limit_buy2 = BuyLimitOrder(trader)
            btc_limit_buy2.update(order_type=TraderOrderType.BUY_LIMIT,
                                  symbol="BTC/USDT",
                                  current_price=decimal.Decimal("10"),
                                  quantity=decimal.Decimal("5000000000"),
                                  price=decimal.Decimal("10"))

            portfolio_manager.portfolio.update_portfolio_available(btc_limit_buy2, True)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.constants as constants
import octobot_trading.enums as enums


class Trade:
    CLOSING_TRADE_ORDER_STATUS = {enums.OrderStatus.CANCELED, enums.OrderStatus.FILLED, enums.OrderStatus.CLOSED}

    def __init__(self, trader):
        self.trader = trader
        self.exchange_manager = trader.exchange_manager

        self.status = enums.OrderStatus.OPEN
        self.creation_time = self.exchange_manager.exchange.get_exchange_current_time()

        self.trade_id = trader.parse_order_id(None)
        # One order might create multiple trades when matched to multiple open orders.
        # Current implementation creates only one trade per order
        # TODO: update this comment when handling multiple trades per order
        self.origin_order_id = None
        self.simulated = True
        self.is_closing_order = False

        self.symbol = None
        self.currency = None
        self.market = None
        self.taker_or_maker = None
        self.origin_price = constants.ZERO
        self.origin_quantity = constants.ZERO
        self.trade_type = None
        self.side = None
        self.executed_quantity = constants.ZERO
        self.canceled_time = 0
        self.executed_time = 0
        self.fee = None
        self.executed_price = constants.ZERO
        self.trade_profitability = constants.ZERO
        self.total_cost = constants.ZERO
        self.reduce_only = False
        self.tag = None
        self.quantity_currency = None

        # raw exchange trade type, used to create trade dict
        self.exchange_trade_type = None

    def update_from_order(self, order, creation_time=0, canceled_time=0, executed_time=0):
        self.currency = order.currency
        self.market = order.market
        self.taker_or_maker = order.taker_or_maker
        self.executed_quantity = order.filled_quantity
        self.executed_price = order.filled_price
        self.origin_price = order.origin_price
        self.origin_quantity = order.origin_quantity
        self.total_cost = order.total_cost
        self.quantity_currency = order.quantity_currency
        self.trade_type = order.order_type
        self.exchange_trade_type = order.exchange_order_type
        self.status = order.status
        self.fee = order.fee
        self.trade_id = order.order_id
        self.origin_order_id = order.order_id
        self.simulated = order.simulated
        self.side = order.side
        self.creation_time = order.creation_time if order.creation_time > 0 else creation_time
        self.canceled_time = order.canceled_time if order.canceled_time > 0 else canceled_time
        self.executed_time = order.executed_time if order.executed_time > 0 else executed_time
        self.symbol = order.symbol
        self.is_closing_order = order.status in self.CLOSING_TRADE_ORDER_STATUS
        self.reduce_only = order.reduce_only
        self.tag = order.tag

    def get_time(self):
        return self.executed_time if self.status is not enums.OrderStatus.CANCELED else self.canceled_time

    def get_quantity(self):
        return self.executed_quantity if self.status is not enums.OrderStatus.CANCELED else self.origin_quantity

    def to_dict(self):
        return {
            enums.ExchangeConstantsOrderColumns.ID.value: self.trade_id,
            enums.ExchangeConstantsOrderColumns.SYMBOL.value: self.symbol,
            enums.ExchangeConstantsOrderColumns.PRICE.value: self.executed_price,
            enums.ExchangeConstantsOrderColumns.STATUS.value: self.status.value,
            enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: self.get_time(),
            enums.ExchangeConstantsOrderColumns.TYPE.value: self.exchange_trade_type.value,
            enums.ExchangeConstantsOrderColumns.SIDE.value: self.side.value,
            enums.ExchangeConstantsOrderColumns.AMOUNT.value: self.get_quantity(),
            enums.ExchangeConstantsOrderColumns.COST.value: self.total_cost,
            enums.ExchangeConstantsOrderColumns.QUANTITY_CURRENCY.value: self.quantity_currency,
            enums.ExchangeConstantsOrderColumns.TAKERORMAKER.value: self.taker_or_maker,
            enums.ExchangeConstantsOrderColumns.FEE.value: self.fee,
            enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value: self.reduce_only,
            enums.ExchangeConstantsOrderColumns.TAG.value: self.tag
        }



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchange_data.contracts.margin_contract as margin_contract


cdef class FutureContract(margin_contract.MarginContract):
    cdef readonly object contract_type
    cdef readonly object position_mode
    cdef public object maintenance_margin_rate

    cdef readonly double minimum_tick_size

    cdef readonly object take_profit_stop_loss_mode

    cpdef bint is_inverse_contract(self)
    cpdef bint is_perpetual_contract(self)
    cpdef bint is_one_way_position_mode(self)
    cpdef void set_position_mode(self, bint is_one_way=*, bint is_hedge=*)
    cpdef void set_take_profit_stop_loss_mode(self, object set_take_profit_stop_loss_mode)
    cpdef object get_fees_currency_side(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.personal_data.orders.order_group as order_group
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.constants as constants
import octobot_commons.logging as logging


class BalancedTakeProfitAndStopOrderGroup(order_group.OrderGroup):
    """
    BalancedTakeProfitAndStopOrderGroup is linking orders together in the way that if any order of the group is filled
    order cancelled, orders that are on the other extreme of the setup (extremes being take profits and stop losses)
    are reduced or cancelled symmetrically.
    Mostly used to pair stop and limit orders together in a stop-loss / take-profit setup using multiple stop losses
    and take profits.
    Orders that are the further from their last_created_price are reduced / cancelled first
    """
    TAKE_PROFIT = "take_profit"
    STOP = "stop"
    UPDATE = "update"
    CANCEL = "cancel"
    ORDER = "order"
    UPDATED_QUANTITY = "updated_quantity"

    def __init__(self, name, orders_manager):
        super().__init__(name, orders_manager)
        # keep track of orders being balanced to avoid nested balance issues
        self.balancing_orders = []

    async def on_fill(self, filled_order, ignored_orders=None):
        """
        Called when an order referencing this group is filled
        This is called right before updating portfolio for this filled order and the
        order fill publication
        :param filled_order: the filled order
        :param ignored_orders: orders that should be ignored
        """
        await self._balance_orders(filled_order, ignored_orders)

    async def on_cancel(self, cancelled_order, ignored_orders=None):
        """
        Called when an order referencing this group is cancelled
        This is called before updating portfolio for this cancelled order and the
        order cancel publication
        :param cancelled_order: the cancelled order
        :param ignored_orders: orders that should be ignored
        """
        await self._balance_orders(cancelled_order, ignored_orders)

    def can_create_order(self, order_type, quantity):
        """
        Returns True if there is an imbalance allowing to create an order of this type and with this quantity
        :param order_type: order type to create
        :param quantity: quantity to put in order
        :return: True when an imbalance allows it
        """
        balance = self._get_balance(None, None)
        if order_util.is_stop_order(order_type):
            return balance[self.STOP].get_balance() + quantity <= balance[self.TAKE_PROFIT].get_balance()
        return balance[self.TAKE_PROFIT].get_balance() + quantity <= balance[self.STOP].get_balance()

    def get_max_order_quantity(self, order_type):
        """
        Returns The maximum order quantity to reach the order side's balance
        :param order_type: order type to create
        :return: the balancing quantity
        """
        balance = self._get_balance(None, None)
        if order_util.is_stop_order(order_type):
            max_quantity = balance[self.TAKE_PROFIT].get_balance() - balance[self.STOP].get_balance()
        else:
            max_quantity = balance[self.STOP].get_balance() - balance[self.TAKE_PROFIT].get_balance()
        return max_quantity if max_quantity > constants.ZERO else constants.ZERO

    async def enable(self, enabled):
        # disable when creating order sequentially (and therefore pause balancing)
        # setting enabled to True will trigger an order balance
        await super().enable(enabled)
        if enabled:
            await self._balance_orders(None, None)

    async def _balance_orders(self, closed_order, ignored_orders):
        if not self.enabled:
            return
        locally_balancing_orders = []
        logger = logging.get_logger(self.__class__.__name__)
        try:
            updated_orders = False
            balance = self._get_balance(closed_order, ignored_orders)
            locally_balancing_orders = balance[self.TAKE_PROFIT].orders + balance[self.STOP].orders
            self.balancing_orders += locally_balancing_orders
            take_profit_actions = balance[self.TAKE_PROFIT].get_actions_to_balance(balance[self.STOP].get_balance())
            stop_actions = balance[self.STOP].get_actions_to_balance(balance[self.TAKE_PROFIT].get_balance())
            for order in take_profit_actions[self.CANCEL] + stop_actions[self.CANCEL]:
                logger.debug(f"Cancelling order to keep balance, order: {order}")
                await order.trader.cancel_order(order, ignored_order=closed_order)
                updated_orders = True
            for update_data in take_profit_actions[self.UPDATE] + stop_actions[self.UPDATE]:
                logger.debug(f"Updating order side to {update_data[self.UPDATED_QUANTITY]} to keep balance, "
                             f"order: {update_data[self.ORDER]}")
                await update_data[self.ORDER].trader.edit_order(
                    update_data[self.ORDER],
                    edited_quantity=update_data[self.UPDATED_QUANTITY])
                updated_orders = True
            if not updated_orders:
                logger.debug("Nothing to update, orders are already evenly balanced")
        except Exception as e:
            logger.exception(e, True, f"Error when balancing orders: {e}")
        finally:
            # remove locally_balancing_orders from self.balancing_orders
            self.balancing_orders = [order
                                     for order in self.balancing_orders
                                     if order not in locally_balancing_orders]

    def _get_balance(self, closed_order, ignored_orders):
        balance = {
            self.TAKE_PROFIT: _SideBalance(),
            self.STOP: _SideBalance()
        }
        for order in self.get_group_open_orders():
            if order is not closed_order \
                    and (ignored_orders is None or order not in ignored_orders) \
                    and order not in self.balancing_orders:
                if order_util.is_stop_order(order.order_type):
                    balance[self.STOP].add_order(order)
                else:
                    balance[self.TAKE_PROFIT].add_order(order)
        return balance


class _SideBalance:
    def __init__(self):
        self.orders = []

    def add_order(self, order):
        self.orders.append(order)
        # orders are sorted according to their distance from their last created price: the further the earlier in list
        # warning: not working with negative prices
        self.orders = sorted(self.orders, key=lambda o: -abs(o.origin_price-o.created_last_price))

    def get_actions_to_balance(self, target_balance):
        actions = {
            BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }
        balance = self.get_balance()
        if target_balance < constants.ZERO:
            target_balance = abs(target_balance)
        if balance <= target_balance:
            # only reducing the current balance is possible
            return actions
        to_be_reduced_amount = constants.ZERO
        remaining_orders = list(self.orders)
        while remaining_orders and balance - to_be_reduced_amount > target_balance:
            order_quantity = remaining_orders[0].origin_quantity
            if balance - to_be_reduced_amount - order_quantity >= target_balance:
                # cancel order and keep reducing
                actions[BalancedTakeProfitAndStopOrderGroup.CANCEL].append(remaining_orders.pop(0))
                to_be_reduced_amount += order_quantity
            else:
                # update order and stop reducing
                actions[BalancedTakeProfitAndStopOrderGroup.UPDATE].append({
                    BalancedTakeProfitAndStopOrderGroup.ORDER: remaining_orders.pop(0),
                    BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY:
                        target_balance - (balance - to_be_reduced_amount - order_quantity)
                })
                break
        return actions

    def get_balance(self):
        if self.orders:
            return sum(order.origin_quantity for order in self.orders)
        return constants.ZERO



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


""" Order class will represent an open order in the specified exchange
In simulation it will also define rules to be filled / canceled
It is also use to store creation & fill values of the order """
cimport octobot_trading.personal_data.portfolios.portfolio as portfolio_class
cimport octobot_trading.personal_data.orders.order as order_class

cdef class SubPortfolio(portfolio_class.Portfolio):
    cdef public portfolio_class.Portfolio parent_portfolio

    cdef public object percent

    cdef public bint is_relative

    cpdef void update_from_parent(self)
    cpdef void set_percent(self, object percent)
    cpdef object update_portfolio_available(self, order_class.Order order, bint is_new_order=*)
    cpdef void reset_portfolio_available(self, str reset_currency=*, object reset_quantity=*)


#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.signals.trading_signal as trading_signal
import octobot_trading.signals.util as signal_util
import octobot_trading.enums as trading_enums
import octobot_trading.errors as trading_errors
import octobot_trading.constants as trading_constants


class SignalBuilder:
    def __init__(self, strategy, exchange, exchange_type, symbol, description, state, orders):
        self.strategy = strategy
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.symbol = symbol
        self.description = description
        self.state = state
        self.orders = orders
        self.reset()

    def add_created_order(self, order, target_amount=None, target_position=None):
        if target_amount is None and target_position is None:
            raise trading_errors.InvalidArgumentError("target_amount or target_position has to be provided")
        if not self._update_pending_orders(order, trading_enums.TradingSignalOrdersActions.CREATE,
                                           target_amount=target_amount,
                                           target_position=target_position):
            self.orders.append(signal_util.create_order_signal_description(
                order,
                trading_enums.TradingSignalOrdersActions.CREATE,
                target_amount=target_amount,
                target_position=target_position
            ))

    def add_order_to_group(self, order):
        if not self._update_pending_orders(order, trading_enums.TradingSignalOrdersActions.ADD_TO_GROUP):
            self.orders.append(signal_util.create_order_signal_description(
                order,
                trading_enums.TradingSignalOrdersActions.ADD_TO_GROUP
            ))

    def add_edited_order(
            self, order,
            updated_target_amount=None,
            updated_target_position=None,
            updated_limit_price=trading_constants.ZERO,
            updated_stop_price=trading_constants.ZERO,
            updated_current_price=trading_constants.ZERO
    ):
        if updated_target_amount is updated_target_position is None and \
           updated_limit_price is updated_stop_price is updated_current_price is trading_constants.ZERO:
            raise trading_errors.InvalidArgumentError("an updated argument has to be provided")
        if not self._update_pending_orders(
                order,
                trading_enums.TradingSignalOrdersActions.EDIT,
                updated_target_amount=updated_target_amount,
                updated_target_position=updated_target_position,
                updated_limit_price=updated_limit_price,
                updated_stop_price=updated_stop_price,
                updated_current_price=updated_current_price
        ):
            self.orders.append(signal_util.create_order_signal_description(
                order,
                trading_enums.TradingSignalOrdersActions.EDIT,
                updated_target_amount=updated_target_amount,
                updated_target_position=updated_target_position,
                updated_limit_price=updated_limit_price,
                updated_stop_price=updated_stop_price,
                updated_current_price=updated_current_price
            ))

    def add_cancelled_order(self, order):
        if not self._update_pending_orders(order, trading_enums.TradingSignalOrdersActions.CANCEL):
            self.orders.append(signal_util.create_order_signal_description(
                order,
                trading_enums.TradingSignalOrdersActions.CANCEL
            ))

    def _update_pending_orders(
        self, order,
        action,
        target_amount=None,
        target_position=None,
        updated_target_amount=None,
        updated_target_position=None,
        updated_limit_price=trading_constants.ZERO,
        updated_stop_price=trading_constants.ZERO,
        updated_current_price=trading_constants.ZERO
    ):
        try:
            index, order_description = self.get_order_description_from_local_orders(order)
            if action is trading_enums.TradingSignalOrdersActions.CREATE:
                # replace order
                self.orders[index] = signal_util.create_order_signal_description(
                    order,
                    trading_enums.TradingSignalOrdersActions.CREATE,
                    target_amount=target_amount,
                    target_position=target_position
                )
            if action is trading_enums.TradingSignalOrdersActions.ADD_TO_GROUP:
                # update order group
                order_description[trading_enums.TradingSignalOrdersAttrs.GROUP_ID.value] = order.order_group.name
                order_description[trading_enums.TradingSignalOrdersAttrs.GROUP_TYPE.value] = \
                    order.order_group.__class__.__name__
            elif action is trading_enums.TradingSignalOrdersActions.EDIT:
                # avoid editing order that are not yet created
                order_description[trading_enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value] = \
                    updated_target_amount
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value] = \
                    updated_target_amount
                order_description[trading_enums.TradingSignalOrdersAttrs.TARGET_POSITION.value] = \
                    updated_target_position
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value] = \
                    updated_target_position
                order_description[trading_enums.TradingSignalOrdersAttrs.LIMIT_PRICE.value] = \
                    float(updated_limit_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value] = \
                    float(updated_limit_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.STOP_PRICE.value] = \
                    float(updated_stop_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value] = \
                    float(updated_stop_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.CURRENT_PRICE.value] = \
                    float(updated_current_price)
                order_description[trading_enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value] = \
                    float(updated_current_price)
            elif action is trading_enums.TradingSignalOrdersActions.CANCEL:
                if order_description[trading_enums.TradingSignalOrdersAttrs.ACTION.value] == \
                   trading_enums.TradingSignalOrdersActions.CREATE.value:
                    # avoid creating order that are to be cancelled
                    self.orders.pop(index)
                else:
                    # now cancel order (no need to perform previous actions as it will get cancelled anyway
                    order_description[trading_enums.TradingSignalOrdersAttrs.ACTION.value] = \
                        trading_enums.TradingSignalOrdersActions.CANCEL.value
            return True
        except trading_errors.OrderDescriptionNotFoundError:
            pass
        return False

    def get_order_description_from_local_orders(self, order):
        for index, order_description in enumerate(self.orders):
            if order_description[trading_enums.TradingSignalOrdersAttrs.SHARED_SIGNAL_ORDER_ID.value] == \
                    order.shared_signal_order_id:
                return index, order_description
        raise trading_errors.OrderDescriptionNotFoundError(f"order not found {order}")

    def build(self):
        return trading_signal.TradingSignal(
            self.strategy,
            self.exchange,
            self.exchange_type,
            self.symbol,
            self.description,
            self.state,
            self.orders
        )

    def reset(self):
        self.orders = []



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import asyncio

import octobot_trading.errors as errors
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.constants as constants
import octobot_trading.exchange_data.ticker.channel.ticker as ticker_channel
import octobot_trading.enums as enums


class TickerUpdater(ticker_channel.TickerProducer):
    CHANNEL_NAME = constants.TICKER_CHANNEL
    TICKER_REFRESH_TIME = 64
    TICKER_FUTURE_REFRESH_TIME = 14
    TICKER_REFRESH_DELAY_THRESHOLD = 10

    def __init__(self, channel):
        super().__init__(channel)
        self._added_pairs = []
        self.is_fetching_future_data = False
        self.refresh_time = self.TICKER_REFRESH_TIME

    async def start(self):
        if self._should_use_future():
            self.is_fetching_future_data = True
            self.refresh_time = self.TICKER_FUTURE_REFRESH_TIME
        if self.channel.is_paused:
            await self.pause()
        else:
            # initialize ticker
            await asyncio.gather(*[self._fetch_ticker(pair)
                                   for pair in self._get_pairs_to_update()])
            await asyncio.sleep(self.refresh_time)
            await self.start_update_loop()

    async def start_update_loop(self):
        while not self.should_stop and not self.channel.is_paused:
            try:
                for pair in self._get_pairs_to_update():
                    await self._fetch_ticker(pair)

                await asyncio.sleep(self.refresh_time)
            except errors.NotSupported:
                self.logger.warning(f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
                await self.pause()
            except Exception as e:
                self.logger.exception(e, True, f"Fail to update ticker : {e}")

    async def _fetch_ticker(self, pair):
        try:
            ticker: dict = await self.channel.exchange_manager.exchange.get_price_ticker(pair)
            if self._is_valid(ticker):
                await self.push(pair, ticker)
                await self.parse_mini_ticker(pair, ticker)
                if self.channel.exchange_manager.is_future:
                    await self.parse_future_data(pair, ticker)
            else:
                self.logger.debug(f"Ignored incomplete ticker: {ticker}")
        except errors.FailedRequest as e:
            self.logger.warning(str(e))

    @staticmethod
    def _is_valid(ticker):
        try:
            # at least require close, volume and timestamp
            return ticker and \
                   all(ticker[field] is not None
                       for field in (
                           enums.ExchangeConstantsTickersColumns.CLOSE.value,
                           enums.ExchangeConstantsTickersColumns.BASE_VOLUME.value,
                           enums.ExchangeConstantsTickersColumns.TIMESTAMP.value
                       ))
        except KeyError:
            return False

    def _cleanup_ticker_dict(self, ticker):
        try:
            ticker.pop("info")
            ticker.pop("symbol")
            ticker.pop("datetime")
        except KeyError as e:
            self.logger.error(f"Fail to cleanup ticker dict ({e})")
        return ticker

    def _get_pairs_to_update(self):
        return self.channel.exchange_manager.exchange_config.traded_symbol_pairs + self._added_pairs

    async def parse_mini_ticker(self, pair, ticker):
        """
        Mini ticker
        """
        try:
            await exchanges_channel.get_chan(constants.MINI_TICKER_CHANNEL,
                                             self.channel.exchange_manager.id).get_internal_producer(). \
                push(symbol=pair, mini_ticker={
                enums.ExchangeConstantsMiniTickerColumns.HIGH_PRICE.value:
                    ticker[enums.ExchangeConstantsTickersColumns.HIGH.value],
                enums.ExchangeConstantsMiniTickerColumns.LOW_PRICE.value:
                    ticker[enums.ExchangeConstantsTickersColumns.LOW.value],
                enums.ExchangeConstantsMiniTickerColumns.OPEN_PRICE.value:
                    ticker[enums.ExchangeConstantsTickersColumns.OPEN.value],
                enums.ExchangeConstantsMiniTickerColumns.CLOSE_PRICE.value:
                    ticker[enums.ExchangeConstantsTickersColumns.CLOSE.value],
                enums.ExchangeConstantsMiniTickerColumns.VOLUME.value:
                    ticker[enums.ExchangeConstantsTickersColumns.BASE_VOLUME.value],
                enums.ExchangeConstantsMiniTickerColumns.TIMESTAMP.value:
                    ticker[enums.ExchangeConstantsTickersColumns.TIMESTAMP.value]})
        except Exception as e:
            self.logger.error(f"Failed to parse mini ticker : {e}")

    """
    Future data management
    """

    def _should_use_future(self):
        return self.channel.exchange_manager.is_future and \
               (
                       self.channel.exchange_manager.exchange.FUNDING_IN_TICKER
                       or self.channel.exchange_manager.exchange.MARK_PRICE_IN_TICKER
               )

    async def parse_future_data(self, symbol: str, ticker: dict):
        if self.channel.exchange_manager.exchange.MARK_PRICE_IN_TICKER:
            await self.extract_mark_price(symbol, ticker)

        if self.channel.exchange_manager.exchange.FUNDING_IN_TICKER:
            await self.extract_funding_rate(symbol, ticker)

    async def extract_mark_price(self, symbol: str, ticker: dict):
        try:
            ticker = self.channel.exchange_manager.exchange.parse_mark_price(ticker, from_ticker=True)
            await exchanges_channel.get_chan(constants.MARK_PRICE_CHANNEL,
                                             self.channel.exchange_manager.id).get_internal_producer(). \
                push(symbol=symbol,
                     mark_price=decimal.Decimal(str(ticker[enums.ExchangeConstantsMarkPriceColumns.MARK_PRICE.value])))
        except Exception as e:
            self.logger.exception(e, True, f"Fail to update mark price from ticker : {e}")

    async def extract_funding_rate(self, symbol: str, ticker: dict):
        try:
            ticker = self.channel.exchange_manager.exchange.parse_funding(ticker, from_ticker=True)
            await exchanges_channel.get_chan(constants.FUNDING_CHANNEL,
                                             self.channel.exchange_manager.id).get_internal_producer(). \
                push(symbol=symbol,
                     funding_rate=decimal.Decimal(str(
                         ticker[enums.ExchangeConstantsFundingColumns.FUNDING_RATE.value])),
                     predicted_funding_rate=decimal.Decimal(str(
                         ticker[enums.ExchangeConstantsFundingColumns.PREDICTED_FUNDING_RATE.value])),
                     next_funding_time=ticker[enums.ExchangeConstantsFundingColumns.NEXT_FUNDING_TIME.value],
                     timestamp=ticker[enums.ExchangeConstantsFundingColumns.LAST_FUNDING_TIME.value])
        except Exception as e:
            self.logger.exception(e, True, f"Fail to update funding rate from ticker : {e}")

    async def modify(self, added_pairs=None, removed_pairs=None):
        if added_pairs:
            to_add_pairs = [pair
                            for pair in added_pairs
                            if pair not in self._get_pairs_to_update()]
            if to_add_pairs:
                self._added_pairs += to_add_pairs
                self.logger.info(f"Added pairs : {to_add_pairs}")
                self._update_refresh_time()

        if removed_pairs:
            self._added_pairs -= removed_pairs
            self.logger.info(f"Removed pairs : {removed_pairs}")
            self._update_refresh_time()

    def _update_refresh_time(self):
        if self.is_fetching_future_data:
            # do not change ticker update rate on futures
            return
        pairs_to_update_count = len(self._get_pairs_to_update())
        delay_multiplier = pairs_to_update_count // self.TICKER_REFRESH_DELAY_THRESHOLD + 1
        # there can be many ticker requests when a large number of currency is in a
        # portfolio, in this case, limit those requests
        self.refresh_time = self.TICKER_REFRESH_TIME * delay_multiplier

    # async def config_callback(self, exchange, cryptocurrency, symbols, time_frames):
    #     if symbols:
    #         await self.modify(added_pairs=symbols)

    async def resume(self) -> None:
        await super().resume()
        if not self.is_running:
            await self.run()



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock

import octobot_trading.exchanges.connectors as exchange_connectors
import octobot_trading.enums as enums
import pytest

from tests.exchanges import exchange_manager

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_initialize_impl_with_none_symbols_and_timeframes(exchange_manager):
    ccxt_exchange = exchange_connectors.CCXTExchange(exchange_manager.config, exchange_manager)

    class MockCCXT:
        def __init__(self):
            self.symbols = None
            self.timeframes = None

        async def load_markets(self):
            pass

        def setSandboxMode(self, is_sandboxed):
            pass

    with mock.patch.object(ccxt_exchange, 'client', new=MockCCXT()):
        await ccxt_exchange.initialize_impl()
        assert ccxt_exchange.symbols == set()
        assert ccxt_exchange.time_frames == set()


async def test_initialize_impl_with_empty_symbols_and_timeframes(exchange_manager):
    ccxt_exchange = exchange_connectors.CCXTExchange(exchange_manager.config, exchange_manager)

    class MockCCXT:
        def __init__(self):
            self.symbols = []
            self.timeframes = []

        async def load_markets(self):
            pass

        def setSandboxMode(self, is_sandboxed):
            pass

    with mock.patch.object(ccxt_exchange, 'client', new=MockCCXT()):
        await ccxt_exchange.initialize_impl()
        assert ccxt_exchange.symbols == set()
        assert ccxt_exchange.time_frames == set()


async def test_initialize_impl(exchange_manager):
    ccxt_exchange = exchange_connectors.CCXTExchange(exchange_manager.config, exchange_manager)

    class MockCCXT:
        def __init__(self):
            self.symbols = [
                "BTC/USDT",
                "ETH/USDT",
                "ETH/BTC",
                "ETH/USDT"
            ]
            self.timeframes = [
                "1h",
                "2h",
                "4h",
                "2h"
            ]

        async def load_markets(self):
            pass

        def setSandboxMode(self, is_sandboxed):
            pass

    with mock.patch.object(ccxt_exchange, 'client', new=MockCCXT()):
        await ccxt_exchange.initialize_impl()
        assert ccxt_exchange.symbols == {
                "BTC/USDT",
                "ETH/USDT",
                "ETH/BTC",
        }
        assert ccxt_exchange.time_frames == {
            "1h",
            "2h",
            "4h",
        }


async def test_set_symbol_partial_take_profit_stop_loss(exchange_manager):
    ccxt_exchange = exchange_connectors.CCXTExchange(exchange_manager.config, exchange_manager)
    with pytest.raises(NotImplementedError):
        await ccxt_exchange.set_symbol_partial_take_profit_stop_loss("BTC/USDT", False,
                                                                     enums.TakeProfitStopLossMode.PARTIAL)


async def test_get_ccxt_order_type(exchange_manager):
    ccxt_exchange = exchange_connectors.CCXTExchange(exchange_manager.config, exchange_manager)
    with pytest.raises(RuntimeError):
        ccxt_exchange.get_ccxt_order_type(None)
    with pytest.raises(RuntimeError):
        ccxt_exchange.get_ccxt_order_type(enums.TraderOrderType.UNKNOWN)
    assert ccxt_exchange.get_ccxt_order_type(enums.TraderOrderType.BUY_LIMIT) == enums.TradeOrderType.LIMIT.value
    assert ccxt_exchange.get_ccxt_order_type(enums.TraderOrderType.STOP_LOSS_LIMIT) == enums.TradeOrderType.LIMIT.value
    assert ccxt_exchange.get_ccxt_order_type(enums.TraderOrderType.TRAILING_STOP) == enums.TradeOrderType.MARKET.value
    assert ccxt_exchange.get_ccxt_order_type(enums.TraderOrderType.SELL_MARKET) == enums.TradeOrderType.MARKET.value



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.databases as databases
import octobot_commons.optimization_campaign as optimization_campaign
import octobot_commons.constants as commons_constants


def get_run_databases_identifier(config, tentacles_setup_config, trading_mode_class=None):
    # avoid importing octobot_trading.modes directly to prevent circular imports
    import octobot_trading.modes as modes

    return databases.RunDatabasesIdentifier(
        trading_mode_class or modes.get_activated_trading_mode(tentacles_setup_config),
        optimization_campaign.OptimizationCampaign.get_campaign_name(tentacles_setup_config),
        backtesting_id=config.get(commons_constants.CONFIG_BACKTESTING_ID),
        optimizer_id=config.get(commons_constants.CONFIG_OPTIMIZER_ID)
    )



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.util as util


cdef class OrderBookManager(util.Initializable):
    cdef object logger

    cdef public bint order_book_initialized

    cdef public double ask_quantity
    cdef public double ask_price
    cdef public double bid_quantity
    cdef public double bid_price

    cdef public object asks # SortedDict
    cdef public object bids # SortedDict

    cdef public double timestamp

    cpdef void reset_order_book(self)
    cpdef void order_book_ticker_update(self, double ask_quantity, double ask_price,
                                        double bid_quantity, double bid_price)
    cpdef void handle_new_book(self, dict orders)
    cpdef void handle_new_books(self, list asks, list bids, object timestamp=*)
    cpdef void handle_book_adds(self, list orders)
    cpdef void handle_book_deletes(self, list orders)
    cpdef void handle_book_updates(self, list orders)
    cpdef tuple get_ask(self)
    cpdef tuple get_bid(self)
    cpdef object get_asks(self, double price)
    cpdef object get_bids(self, double price)

    cdef object _handle_book_delete(self, dict order) # using object to prevent ignoring KeyError
    cdef object _handle_book_update(self, dict order) # using object to prevent ignoring KeyError
    cdef object _handle_book_add(self, dict order) # using object to prevent ignoring KeyError
    cdef void _set_asks(self, double price, list asks)
    cdef void _set_bids(self, double price, list bids)
    cdef void _remove_asks(self, double price)
    cdef void _remove_bids(self, double price)

cdef int _order_id_index(str order_id, list order_list)
cdef list _convert_price_size_list_to_order(list price_size_list, str side)
cdef dict _convert_price_size_to_order(list price_size, str side)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import trading_backend

import octobot_commons.logging as logging
import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.tentacles_management as tentacles_management

import octobot_tentacles_manager.api as api

import octobot_trading.enums as enums
import octobot_trading.exchanges.types as exchanges_types
import octobot_trading.exchanges.exchange_manager as exchange_manager
import octobot_trading.exchanges.exchange_factory as exchange_factory


def get_margin_exchange_class(exchange_name, tentacles_setup_config):
    return search_exchange_class_from_exchange_name(exchanges_types.MarginExchange, exchange_name,
                                                    tentacles_setup_config)


def get_future_exchange_class(exchange_name, tentacles_setup_config):
    return search_exchange_class_from_exchange_name(exchanges_types.FutureExchange, exchange_name,
                                                    tentacles_setup_config)


def get_spot_exchange_class(exchange_name, tentacles_setup_config):
    return search_exchange_class_from_exchange_name(exchanges_types.SpotExchange, exchange_name,
                                                    tentacles_setup_config)


def search_exchange_class_from_exchange_name(exchange_class, exchange_name,
                                             tentacles_setup_config, enable_default=False):
    for exchange_candidate in tentacles_management.get_all_classes_from_parent(exchange_class):
        try:
            if _is_exchange_candidate_matching(exchange_candidate, exchange_name,
                                               tentacles_setup_config, enable_default=enable_default):
                return exchange_candidate
        except NotImplementedError:
            # A subclass of AbstractExchange will raise a NotImplementedError when calling its get_name() method
            # Here we are returning only a subclass that matches the expected name
            # Only Exchange Tentacles are implementing get_name() to specify the related exchange
            # As we are searching for an exchange_type specific subclass
            # We should ignore classes that raises NotImplementedError
            pass

    if enable_default:
        return None

    logging.get_logger().debug(f"No specific exchange implementation for {exchange_name} found, using a default one.")
    # TODO: handle default future exchange instead of creating a SpotExchange
    return search_exchange_class_from_exchange_name(exchanges_types.SpotExchange, exchange_name,
                                                    tentacles_setup_config, enable_default=True)


def _is_exchange_candidate_matching(exchange_candidate, exchange_name, tentacles_setup_config, enable_default=False):
    return not exchange_candidate.is_simulated_exchange() and \
           (not exchange_candidate.is_default_exchange() or enable_default) and \
           exchange_candidate.is_supporting_exchange(exchange_name) and \
           (tentacles_setup_config is None or
            api.is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config,
                                                                exchange_candidate.__name__))


def get_order_side(order_type):
    return enums.TradeOrderSide.BUY.value if order_type in (enums.TraderOrderType.BUY_LIMIT,
                                                            enums.TraderOrderType.BUY_MARKET) \
        else enums.TradeOrderSide.SELL.value


def log_time_sync_error(logger, exchange_name, error, details):
    logger.error(
        f"{_get_time_sync_error_message(exchange_name, details)} Error: {error}")


def _get_docs_url():
    try:
        import octobot.constants
        return octobot.constants.OCTOBOT_DOCS_URL
    except ImportError:
        return "https://docs.octobot.online"


def _get_exchanges_docs_url():
    try:
        import octobot.constants
        return octobot.constants.EXCHANGES_DOCS_URL
    except ImportError:
        return "https://exchanges.docs.octobot.online"


def _get_time_sync_error_message(exchange_name, details):
    return f"Time synchronization error when loading your {exchange_name.capitalize()} {details}. " \
        f"To fix this, please synchronize your computer's clock. See " \
        f"{_get_docs_url()}/installation/installation-troubleshoot#time-synchronization"


def get_partners_explanation_message():
    return f"More info on partner exchanges on {_get_exchanges_docs_url()}#partner-exchanges-support-octobot"


def _get_minimal_exchange_config(exchange_name, exchange_config):
    return {
        common_constants.CONFIG_EXCHANGES: {
            exchange_name: exchange_config
        },
        common_constants.CONFIG_TRADER: {
            common_constants.CONFIG_ENABLED_OPTION: False
        },
        common_constants.CONFIG_SIMULATOR: {
            common_constants.CONFIG_ENABLED_OPTION: False
        }
    }


async def is_compatible_account(exchange_name: str, exchange_config: dict, tentacles_setup_config) -> (bool, str):
    local_exchange_manager = exchange_manager.ExchangeManager(
        _get_minimal_exchange_config(exchange_name, exchange_config),
        exchange_name
    )
    local_exchange_manager.tentacles_setup_config = tentacles_setup_config
    # Avoid exchange authentication error log
    local_exchange_manager.is_collecting = True
    await exchange_factory.search_and_create_spot_exchange(local_exchange_manager)
    backend = trading_backend.exchange_factory.create_exchange_backend(local_exchange_manager.exchange)
    try:
        is_compatible, error = await backend.is_valid_account()
        message = f"[REST API] {exchange_name.capitalize()} connection can be faster " \
                  f"using websockets. {error}. " \
                  f"Please create a new {exchange_name.capitalize()} account to use websockets. "
        return is_compatible, message if error else error
    except trading_backend.TimeSyncError:
        return False, _get_time_sync_error_message(exchange_name, "account details")
    except trading_backend.ExchangeAuthError:
        return False, f"Invalid {exchange_name.capitalize()} authentication details"
    except Exception as e:
        return False, f"Error when loading exchange account: {e}"
    finally:
        # do not log stopping message
        local_exchange_manager.exchange.connector.logger.disable(True)
        await local_exchange_manager.exchange.stop()
        local_exchange_manager.exchange.connector.logger.disable(False)


async def get_historical_ohlcv(local_exchange_manager, symbol, time_frame, start_time, end_time):
    """
    Async generator, use as follows:
        async for candles in get_historical_ohlcv(exchange_manager, pair, time_frame, start_time, end_time):
            # candles stuff
    """
    reached_max = False
    while start_time < end_time and not reached_max:
        candles = await local_exchange_manager.exchange.get_symbol_prices(symbol, time_frame, since=int(start_time))
        if candles:
            start_time = candles[-1][common_enums.PriceIndexes.IND_PRICE_TIME.value]
            while start_time > end_time and candles:
                start_time = candles.pop(-1)[common_enums.PriceIndexes.IND_PRICE_TIME.value]
                reached_max = True
            if candles:
                local_exchange_manager.exchange.uniformize_candles_if_necessary(candles)
                yield candles
                # avoid fetching the last element twice
                start_time += 1
            else:
                reached_max = True
        else:
            reached_max = True


def get_exchange_type(exchange_manager_instance):
    if exchange_manager_instance.is_spot_only:
        return enums.ExchangeTypes.SPOT
    if exchange_manager_instance.is_future:
        return enums.ExchangeTypes.FUTURE
    if exchange_manager_instance.is_margin:
        return enums.ExchangeTypes.MARGIN
    return enums.ExchangeTypes.SPOT



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport numpy as np
np.import_array()

cimport octobot_trading.exchange_data.exchange_symbol_data as exchange_symbol_data

cpdef np.ndarray get_symbol_close_candles(exchange_symbol_data.ExchangeSymbolData symbol_data, str time_frame,  int limit, bint include_in_construction)
cpdef np.ndarray get_symbol_open_candles(exchange_symbol_data.ExchangeSymbolData symbol_data, str time_frame, int limit, bint include_in_construction)
cpdef np.ndarray get_symbol_high_candles(exchange_symbol_data.ExchangeSymbolData symbol_data, str time_frame, int limit, bint include_in_construction)
cpdef np.ndarray get_symbol_low_candles(exchange_symbol_data.ExchangeSymbolData symbol_data, str time_frame, int limit, bint include_in_construction)
cpdef np.ndarray get_symbol_volume_candles(exchange_symbol_data.ExchangeSymbolData symbol_data, str time_frame, int limit, bint include_in_construction)
cpdef np.ndarray get_symbol_time_candles(exchange_symbol_data.ExchangeSymbolData symbol_data, str time_frame, int limit, bint include_in_construction)

cdef np.ndarray _add_in_construction_data(np.ndarray candles, exchange_symbol_data.ExchangeSymbolData symbol_data, object time_frame, int data_type)


#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.constants as constants
import octobot_trading.modes.script_keywords.basic_keywords.user_inputs as user_inputs
import octobot_trading.enums as enums
import octobot_trading.errors as errors


async def user_select_leverage(
        ctx,
        def_val=1,
        order=None,
        name="leverage"):
    return await user_inputs.user_input(ctx, name, "int", def_val, order=order)


async def user_select_emit_trading_signals(ctx, identifier, def_val=False):
    await user_inputs.user_input(ctx, constants.CONFIG_TRADING_SIGNALS_STRATEGY, "text", identifier,
                                 show_in_summary=False, show_in_optimizer=False)
    return await user_inputs.user_input(ctx, "emit_trading_signals", "boolean", def_val,
                                        show_in_summary=False, show_in_optimizer=False)


async def set_leverage(ctx, leverage):
    if ctx.exchange_manager.is_future:
        side = None
        try:
            await ctx.exchange_manager.trader.set_leverage(ctx.symbol, side, decimal.Decimal(str(leverage)))
        except errors.ContractExistsError as e:
            ctx.logger.debug(str(e))
        except NotImplementedError as e:
            ctx.logger.exception(e, True, str(e))
        except AttributeError:
            # TODO remove this except when bybit tentacle is up
            ctx.logger.warning("TODO: rebase tentacles when bybit exchange is up")
        except Exception as e:
            ctx.logger.exception(e, True, str(e))


async def set_partial_take_profit_stop_loss(ctx, tp_sl_mode=enums.TakeProfitStopLossMode.PARTIAL.value):
    if ctx.exchange_manager.is_future:
        await ctx.exchange_manager.trader.set_symbol_take_profit_stop_loss_mode(
            ctx.symbol, enums.TakeProfitStopLossMode(tp_sl_mode)
        )



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


""" Order class will represent an open order in the specified exchange
In simulation it will also define rules to be filled / canceled
It is also use to store creation & fill values of the order """
cimport octobot_trading.util as util
cimport octobot_trading.personal_data.orders.order_state as orders_states

cdef class Order(util.Initializable):
    cdef public object trader
    cdef public object exchange_manager

    cdef public object side # TradeOrderSide
    cdef public object status # OrderStatus
    cdef public object order_type # TraderOrderType
    cdef public object lock # Lock

    cdef public orders_states.OrderState state

    cdef public bint is_synchronized_with_exchange
    cdef public bint is_from_this_octobot
    cdef public bint simulated
    cdef public bint reduce_only
    cdef public bint close_position

    cdef public bint created
    cdef public str symbol
    cdef public str currency
    cdef public str market
    cdef public str quantity_currency
    cdef public str taker_or_maker
    cdef public str order_id
    cdef public str logger_name
    cdef public str tag

    cdef readonly str shared_signal_order_id

    cdef public object origin_price
    cdef public object origin_stop_price
    cdef public object origin_quantity
    cdef public object filled_quantity
    cdef public object filled_price
    cdef public object total_cost
    cdef public object created_last_price
    cdef public object order_profitability
    cdef public object position_side

    cdef public object order_group

    cdef public double timestamp
    cdef public double creation_time
    cdef public double canceled_time
    cdef public double executed_time

    cdef public dict fee # Dict[str, Union[str, decimal.Decimal]]
    cdef public object fees_currency_side   # trading_enums.FeesCurrencySide

    cdef list last_prices
    cdef public list chained_orders # List[Order]
    cdef public object triggered_by # Order
    cdef public bint has_been_bundled
    cdef public bint is_waiting_for_chained_trigger
    cdef public dict exchange_creation_params
    cdef public dict trader_creation_kwargs


    cdef public object exchange_order_type # raw exchange order type, used to create order dict

    cpdef bint update(self,
            str symbol,
            str order_id=*,
            object status=*,
            object current_price=*,
            object quantity=*,
            object price=*,
            object stop_price=*,
            object quantity_filled=*,
            object filled_price=*,
            object average_price=*,
            dict fee=*,
            object total_cost=*,
            object timestamp=*,
            object order_type=*,
            object reduce_only=*,
            object close_position=*,
            object position_side=*,
            object fees_currency_side=*,
            object group=*,
            str tag=*,
            str quantity_currency=*)
    cdef object _update_type_from_raw(self, dict raw_order)  # return object to allow exception raising
    cdef void _update_taker_maker(self)
    cdef object _on_origin_price_change(self, object previous_price, object price_time)

    cpdef str to_string(self)
    cpdef object get_total_fees(self, str currency)
    cpdef bint is_created(self)
    cpdef bint is_open(self)
    cpdef bint is_filled(self)
    cpdef bint is_cancelled(self)
    cpdef bint is_closed(self)
    cpdef bint is_long(self)
    cpdef bint is_short(self)
    cpdef bint is_refreshing(self)
    cpdef bint can_be_edited(self)
    cpdef object get_position_side(self, object future_contract)
    cpdef void on_fill_actions(self)
    cpdef dict get_computed_fee(self, object forced_value=*)
    cpdef object get_profitability(self)
    cpdef double generate_executed_time(self)
    cpdef bint is_counted_in_available_funds(self)
    cpdef bint is_self_managed(self)
    cpdef object update_from_raw(self, dict raw_order)
    cpdef void consider_as_filled(self)
    cpdef void consider_as_canceled(self)
    cpdef dict to_dict(self)
    cpdef void clear(self)
    cpdef bint is_to_be_maintained(self)
    cpdef str get_logger_name(self)
    cpdef void add_chained_order(self, object chained_order)
    cpdef bint should_be_created(self)
    cpdef void add_to_order_group(self, object order_group)
    cdef void _update_total_cost(self)

cdef object _get_sell_and_buy_types(object order_type)
cdef object _infer_order_type_from_maker_or_taker(dict raw_order, object side)

cpdef tuple parse_order_type(dict raw_order)
